"""The Mimeo File Consumer module.

It exports only one class:
    * FileConsumer
        A Consumer implementation saving data in the filesystem.
"""
import logging
from pathlib import Path

from mimeo.config.mimeo_config import MimeoOutputDetails
from mimeo.consumers import Consumer

logger = logging.getLogger(__name__)


class FileConsumer(Consumer):
    """A Consumer implementation saving data in the filesystem.

    This Consumer is instantiated for the 'file' output direction
    and saves data produced by Mimeo as files using Mimeo Output Details.

    Methods
    -------
    consume
        Save data generated by Mimeo into a file.

    Attributes
    ----------
    directory : str
        A directory path to save files within
    output_path_tmplt : str
        An output file path template
        (every file has its index inside the actual path)
    """

    def __init__(self, output_details: MimeoOutputDetails):
        """Initialize FileConsumer class.

        Parameters
        ----------
        output_details : MimeoOutputDetails
            Configured Mimeo Output Details
        """
        self.directory = output_details.directory_path
        self.output_path_tmplt = f"{self.directory}/{output_details.file_name_tmplt}"
        self.__count = 0

    def consume(self, data: str) -> None:
        """Save data generated by Mimeo into a file.

        It is an implementation of Consumer's abstract method.
        If the output directory does not exist it is created.
        Every file name has an index inside its path.

        Parameters
        ----------
        data : str
            Stringified data generated by Mimeo
        """
        logger.fine(f"Consuming data [{data}]")
        if not Path(self.directory).exists():
            logger.info(f"Creating output directory [{self.directory}]")
            Path(self.directory).mkdir(parents=True, exist_ok=True)

        self.__count += 1
        file_name = self.output_path_tmplt.format(self.__count)

        logger.info(f"Writing data into file [{file_name}]")
        with open(file_name, "w") as output_file:
            output_file.write(data)



"""The Mimeo Tools module.

It is meant to store all useful functions that can help in any
non-Mimeo-specific operations. It exports the following functions:
    * get_resource(resource_name: str) -> TextIO
        Return a Mimeo resource.
"""
from typing.io import TextIO

from mimeo.resources.exc import ResourceNotFound

try:
    import importlib.resources as pkg_resources
except ImportError:  # Python < 3.7
    import importlib_resources as pkg_resources

from mimeo import resources as data


def get_resource(resource_name: str) -> TextIO:
    """Return a Mimeo resource.

    The resource needs to be included in mimeo.resources package
    to be returned.

    Parameters
    ----------
    resource_name : str
        A Mimeo resource name

    Returns
    -------
    TextIO
        A Mimeo resource

    Raises
    ------
    ResourceNotFound
        If the resource does not exist
    """
    try:
        return pkg_resources.open_text(data, resource_name)
    except FileNotFoundError:
        raise ResourceNotFound(resource_name)



"""The Mimeo Configuration package.

It contains the following modules:
* mimeo_config
    The Mimeo Configuration module
* exc
    The Mimeo Configuration Exceptions module

The Mimeo Configuration package exports a class representing
root Mimeo Configuration component:
* MimeoConfig
    A MimeoDTO class representing Mimeo Configuration

To use this package, simply import it:
    from mimeo.config import MimeoConfig
    from mimeo.config.exc import UnsupportedPropertyValue
"""
from .mimeo_config import MimeoConfig



"""The Mimeo Utils Exceptions module.

It contains all custom exceptions related to Mimeo Utils:
    * InvalidMimeoUtil
        A custom Exception class for an invalid Mimeo Util.
    * InvalidValue
        A custom Exception class for an invalid value in Mimeo Util.
    * NotASpecialField
        A custom Exception class for a field used as a special.
"""


class InvalidMimeoUtil(Exception):
    """A custom Exception class for an invalid Mimeo Util.

    Raised when Mimeo Util node has missing _name property, or it
    does not match any Mimeo Util.
    """

    pass


class InvalidValue(Exception):
    """A custom Exception class for an invalid value in Mimeo Util.

    Raised when Mimeo Util node is incorrectly parametrized.
    """

    pass


class NotASpecialField(Exception):
    """A custom Exception class for a field used as a special.

    Raised while attempting to retrieve special field name when it is
    not a special one.
    """

    def __init__(self, field_name: str):
        """Initialize NotASpecialField exception with details.

        Extends Exception constructor with a custom message.

        Parameters
        ----------
        field_name : str
            A field name
        """
        super().__init__(f"Provided field [{field_name}] is not a special one (use {'{:NAME:}'})!")



"""The Mimeo Database Exceptions module.

It contains all custom exceptions related to Mimeo Context:
    * InvalidIndex
        A custom Exception class for invalid row index.
    * InvalidSex
        A custom Exception class for uninitialized context's iteration.
    * CountryNotFound
        A custom Exception class for not found context's iteration.
    * OutOfStock
        A custom Exception class for invalid special field's name.
"""


class InvalidIndex(Exception):
    """A custom Exception class for invalid row index.

    Raised while attempting to get data at row that does not exist.
    """

    def __init__(self, index: int, last_index: int):
        """Initialize InvalidIndex exception with details.

        Extends Exception constructor with a custom message.

        Parameters
        ----------
        index : int
            An invalid index
        last_index : str
            The last existing index
        """
        super().__init__(f"Provided index [{index}] is out or the range: 0-{last_index}!")


class InvalidSex(Exception):
    """A custom Exception class for invalid sex.

    Raised when sex provided to filter forenames is not supported.
    """

    def __init__(self, supported_sex_list: tuple):
        """Initialize InvalidSex exception with details.

        Extends Exception constructor with a custom message.

        Parameters
        ----------
        supported_sex_list : int
            A list of supported sex values
        """
        super().__init__(f"Invalid sex (use {' / '.join(supported_sex_list)})!")


class DataNotFound(Exception):
    """A custom Exception class for not found data.

    Raised while attempting to filter data by a column's value that
    does not match any row.
    """

    pass


class OutOfStock(Exception):
    """A custom Exception class for consuming all data.

    Raised while attempting to get next unique value when all were
    consumed already.
    """

    pass



"""The Mimeo Context Manager module.

It exports only one class:
    * MimeoContextManager
        An OnlyOneAlive class managing Mimeo Contexts.
"""
from __future__ import annotations

from typing import Union

from mimeo.config import MimeoConfig
from mimeo.context import MimeoContext
from mimeo.context.exc import VarNotFound
from mimeo.meta import Alive, OnlyOneAlive


class MimeoContextManager(Alive, metaclass=OnlyOneAlive):
    """An OnlyOneAlive class managing Mimeo Contexts.

    It allows you to initialize a context, get the currently processing
    context, switch it or reach any other. Besides that it gives you
    access to Mimeo Vars.
    The only way to use it successfully it is by `with` statement:
        with MimeoContextManager(mimeo_config) as mimeo_mng:
            ...

    Methods
    -------
    get_context(self, context: str) -> MimeoContext
        Return a Mimeo Context with a specific name.
    get_current_context(self) -> MimeoContext
        Return the current Mimeo Context.
    set_current_context(self, context: MimeoContext)
        Set the current Mimeo Context.
    get_var(self, variable_name: str)
        Return a specific Mimeo Var value.
    """

    def __init__(self, mimeo_config: MimeoConfig = None):
        """Initialize MimeoContextManager class.

        Parameters
        ----------
        mimeo_config : MimeoConfig
            The Mimeo Configuration
        """
        super().__init__()
        self._mimeo_config = mimeo_config
        self._vars = {}
        self._contexts = {}
        self._current_context = None

    def __enter__(self) -> MimeoContextManager:
        """Enter the MimeoContextManager instance.

        Extends Alive __enter__ function and initializes vars.

        Returns
        -------
        self : MimeoContextManager
            A MimeoContextManager instance
        """
        super().__enter__()
        self._vars = self._mimeo_config.vars
        return self

    def __exit__(self, exc_type, exc_val, exc_tb) -> None:
        """Exit the MimeoContextManager instance.

        Extends Alive __enter__ function and removes internal
        attributes.

        Parameters
        ----------
        exc_type
            An exception's type
        exc_val
            An exception's value
        exc_tb
            An exception's traceback

        Returns
        -------
        None
            A None value
        """
        super().__exit__(exc_type, exc_val, exc_tb)
        self._vars = None
        self._contexts = None
        return None

    def get_context(self, context: str) -> MimeoContext:
        """Return a Mimeo Context with a specific name.

        If the context does not exist, it is initialized.

        Parameters
        ----------
        context : str
            A context's name

        Returns
        -------
        MimeoContext
            A specific Mimeo Context

        Raises
        ------
        InstanceNotAlive
            If the MimeoContextManager instance is not alive
        """
        super().assert_alive()
        if context not in self._contexts:
            self._contexts[context] = MimeoContext(context)
        return self._contexts[context]

    def get_current_context(self) -> MimeoContext:
        """Return the current Mimeo Context.

        Returns
        -------
        MimeoContext
            The current Mimeo Context

        Raises
        ------
        InstanceNotAlive
            If the MimeoContextManager instance is not alive
        """
        super().assert_alive()
        return self._current_context

    def set_current_context(self, context: MimeoContext):
        """Set the current Mimeo Context.

        Parameters
        ----------
        context : MimeoContext
            A Mimeo Context to be set as the current one

        Raises
        ------
        InstanceNotAlive
            If the MimeoContextManager instance is not alive
        """
        super().assert_alive()
        self._current_context = context

    def get_var(self, variable_name: str) -> Union[str, int, bool, dict]:
        """Return a specific Mimeo Var value.

        Parameters
        ----------
        variable_name : str
            The Mimeo Var name

        Returns
        -------
        value : Union[str, int, bool, dict]
            The Mimeo Var value

        Raises
        ------
        InstanceNotAlive
            If the MimeoContextManager instance is not alive
        VarNotFound
            If the Mimeo Var with the `variable_name` provided does not
            exist
        """
        super().assert_alive()
        value = self._vars.get(variable_name)
        if value is not None:
            return value
        else:
            raise VarNotFound(variable_name)



"""The Mimeo Generator Factory module.

It exports only one class:
    * GeneratorFactory
        A Factory class instantiating a Generator based on Mimeo Config.
"""
from mimeo.config.exc import UnsupportedPropertyValue
from mimeo.config.mimeo_config import MimeoConfig
from mimeo.generators import Generator, XMLGenerator


class GeneratorFactory:
    """A Factory class instantiating a Generator based on Mimeo Config.

    Implementation of the Generator class depends on the output format
    configured.

    Attributes
    ----------
    XML
        The 'xml' output format

    Methods
    -------
    get_generator(mimeo_config: MimeoConfig) -> Generator
        Initialize a Generator based on the Mimeo Output Format.
    """

    XML = MimeoConfig.OUTPUT_FORMAT_XML

    @staticmethod
    def get_generator(mimeo_config: MimeoConfig) -> Generator:
        """Initialize a Generator based on the Mimeo Output Format.

        Parameters
        ----------
        mimeo_config : MimeoConfig
            A Mimeo Configuration

        Returns
        -------
        Generator
            A Generator's implementation instance

        Raises
        ------
        UnsupportedPropertyValue
            If the output format is not supported
        """
        output_format = mimeo_config.output_details.format
        if output_format == GeneratorFactory.XML:
            return XMLGenerator(mimeo_config)
        else:
            raise UnsupportedPropertyValue(MimeoConfig.OUTPUT_DETAILS_FORMAT_KEY,
                                           output_format,
                                           MimeoConfig.SUPPORTED_OUTPUT_FORMATS)



"""The Mimeo module.

This module is a main module providing the most high level class
in Mimeo:
    * Mimeograph
        A class responsible for the Mimeo processing.
"""
import logging

from mimeo.config.mimeo_config import MimeoConfig
from mimeo.consumers import ConsumerFactory
from mimeo.context import MimeoContextManager
from mimeo.generators import GeneratorFactory

logger = logging.getLogger(__name__)


class Mimeograph:
    """A class responsible for the Mimeo processing.

    Based on the Mimeo Configuration it instantiates generator and
    consumer to produce desired data.

    Methods
    -------
    process()
        Process the Mimeo Configuration (generate data and consume).
    """

    def __init__(self, mimeo_config: MimeoConfig):
        self._mimeo_config = mimeo_config
        self._generator = GeneratorFactory.get_generator(self._mimeo_config)
        self._consumer = ConsumerFactory.get_consumer(self._mimeo_config)

    def process(self):
        """Process the Mimeo Configuration (generate data and consume)."""
        logger.info("Starting data generation")
        with MimeoContextManager(self._mimeo_config):
            for data in self._generator.generate(self._mimeo_config.templates):
                data_str = self._generator.stringify(data, self._mimeo_config)
                self._consumer.consume(data_str)
        logger.info("Data has been processed")



"""The Cities module.

It exports classes related to cities CSV data:
    * City
        DTO class representing a single row in cities CSV data.
    * CitiesDB
        Class exposing READ operations on cities CSV data.
"""
from typing import List

import pandas

from mimeo import tools
from mimeo.database.exc import InvalidIndex


class City:
    """DTO class representing a single row in cities CSV data.

    Attributes
    ----------
    id : int
        An identifier
    name : str
        A city name
    name_ascii : str
        A city name in ASCII encoding
    country : str
        A country of a city
    """

    def __init__(self, identifier: str, name: str, name_ascii: str, country: str):
        """Initialize City class.

        Parameters
        ----------
        identifier : str
            An identifier
        name : str
            A city name
        name_ascii : str
            A city name in ASCII encoding
        country : str
            A country of a city
        """
        self.id = int(identifier)
        self.name = name
        self.name_ascii = name_ascii
        self.country = country

    def __str__(self) -> str:
        """Stringify the City instance.

        Returns
        -------
        str
            A stringified `dict` representation of the City instance
        """
        return str({
            "id": self.id,
            "name": self.name,
            "name_ascii": self.name_ascii,
            "country": self.country
        })

    def __repr__(self) -> str:
        """Represent the City instance.

        Returns
        -------
        str
            A python representation of the City instance
        """
        return f"City('{self.id}', '{self.name}', '{self.name_ascii}', '{self.country}')"


class CitiesDB:
    """Class exposing READ operations on cities CSV data.

    Attributes
    ----------
    NUM_OF_RECORDS : int
        A number of rows in cities CSV data

    Methods
    -------
    get_cities() -> List[City]
        Get all cities.
    get_cities_of(country_iso3: str) -> List[City]
        Get cities of a specific country.
    get_city_at(index: int) -> City
        Get a city at `index` position.
    """

    NUM_OF_RECORDS = 42905
    _CITIES_DB = "cities.csv"
    _CITIES_DF = None
    _CITIES = None
    _COUNTRY_CITIES = {}

    def get_city_at(self, index: int) -> City:
        """Get a city at `index` position.

        Parameters
        ----------
        index : int
            A city row index

        Returns
        -------
        City
            A specific city

        Raises
        ------
        InvalidIndex
            If the provided `index` is out of bounds
        """
        cities = self._get_cities()
        try:
            return cities[index]
        except IndexError:
            raise InvalidIndex(index, CitiesDB.NUM_OF_RECORDS-1)

    def get_cities_of(self, country_iso3: str) -> List[City]:
        """Get cities of a specific country.

        Parameters
        ----------
        country_iso3 : str
            A country ISO3 code to filter cities

        Returns
        -------
        List[City]
            List of cities filtered by country
        """
        return self._get_country_cities(country_iso3).copy()

    @classmethod
    def get_cities(cls) -> List[City]:
        """Get all cities.

        Returns
        -------
        List[City]
            List of all cities
        """
        return cls._get_cities().copy()

    @classmethod
    def _get_country_cities(cls, country_iso3: str) -> List[City]:
        """Get cities of a specific country from cache.

        The country's cities list is initialized for the first time
        and cached in internal class attribute.

        Parameters
        ----------
        country_iso3 : str
            A country ISO3 code to filter cities

        Returns
        -------
        List[City]
            List of cities filtered by country
        """
        if country_iso3 not in cls._COUNTRY_CITIES:
            cities = cls._get_cities()
            cls._COUNTRY_CITIES[country_iso3] = list(filter(lambda city: city.country == country_iso3, cities))
        return cls._COUNTRY_CITIES[country_iso3]

    @classmethod
    def _get_cities(cls) -> List[City]:
        """Get all cities from cache.

        The cities list is initialized for the first time and cached
        in internal class attribute.

        Returns
        -------
        List[City]
            List of all cities
        """
        if cls._CITIES is None:
            cls._CITIES = [City(row.ID, row.CITY, row.CITY_ASCII, row.COUNTRY)
                           for row in cls._get_cities_df().itertuples()]
        return cls._CITIES

    @classmethod
    def _get_cities_df(cls) -> pandas.DataFrame:
        """Load cities CSV data and save in internal class attribute."""
        if cls._CITIES_DF is None:
            cls._CITIES_DF = pandas.read_csv(tools.get_resource(CitiesDB._CITIES_DB))
        return cls._CITIES_DF



"""The Countries module.

It exports classes related to countries CSV data:
    * Country
        DTO class representing a single row in countries CSV
        data.
    * CountriesDB
        Class exposing READ operations on countries CSV data.
"""
from typing import List

import pandas

from mimeo import tools
from mimeo.database.exc import InvalidIndex


class Country:
    """DTO class representing a single row in cities CSV data.

    Attributes
    ----------
    iso_3 : str
        A country ISO3 code
    iso_2 : str
        A country ISO2 code
    name : str
        A country name
    """

    def __init__(self, iso_3: str, iso_2: str, name: str):
        """Initialize Country class.

        Parameters
        ----------
        iso_3 : str
            A country ISO3 code
        iso_2 : str
            A country ISO2 code
        name : str
            A country name
        """
        self.iso_3 = iso_3
        self.iso_2 = iso_2
        self.name = name

    def __str__(self) -> str:
        """Stringify the Country instance.

        Returns
        -------
        str
            A stringified `dict` representation of the Country instance
        """
        return str({
            "iso_3": self.iso_3,
            "iso_2": self.iso_2,
            "name": self.name
        })

    def __repr__(self) -> str:
        """Represent the Country instance.

        Returns
        -------
        str
            A python representation of the Country instance
        """
        return f"Country('{self.iso_3}', '{self.iso_2}', '{self.name}')"


class CountriesDB:
    """Class exposing READ operations on countries CSV data.

    Attributes
    ----------
    NUM_OF_RECORDS : int
        A number of rows in countries CSV data

    Methods
    -------
    get_countries() -> List[Country]
        Get all countries.
    get_country_at(index: int) -> Country
        Get a country at `index` position.
    get_country_by_iso_3(iso_3: str) -> Country
        Get a country having a specific ISO3 code.
    get_country_by_iso_3(iso_2: str) -> Country
        Get a country having a specific ISO2 code.
    get_country_by_name(name: str) -> Country
        Get a country having a specific name.
    """

    NUM_OF_RECORDS = 239
    _COUNTRIES_DB = "countries.csv"
    _COUNTRIES_DF = None
    _COUNTRIES = None

    def get_country_at(self, index: int) -> Country:
        """Get a country at `index` position.

        Parameters
        ----------
        index : int
            A country row index

        Returns
        -------
        Country
            A specific country

        Raises
        ------
        InvalidIndex
            If the provided `index` is out of bounds
        """
        countries = self._get_countries()
        try:
            return countries[index]
        except IndexError:
            raise InvalidIndex(index, CountriesDB.NUM_OF_RECORDS-1)

    def get_country_by_iso_3(self, iso_3: str) -> Country:
        """Get a country having a specific ISO3 code.

        Parameters
        ----------
        iso_3 : str
            An ISO3 code to find a country

        Returns
        -------
        Country
            A specific country or None
        """
        countries = self._get_countries()
        return next(filter(lambda country: country.iso_3 == iso_3, countries), None)

    def get_country_by_iso_2(self, iso_2: str) -> Country:
        """Get a country having a specific ISO2 code.

        Parameters
        ----------
        iso_2 : str
            An ISO2 code to find a country

        Returns
        -------
        Country
            A specific country or None
        """
        countries = self._get_countries()
        return next(filter(lambda country: country.iso_2 == iso_2, countries), None)

    def get_country_by_name(self, name: str) -> Country:
        """Get a country having a specific name.

        Parameters
        ----------
        name : str
            A name to find a country

        Returns
        -------
        Country
            A specific country or None
        """
        countries = self._get_countries()
        return next(filter(lambda country: country.name == name, countries), None)

    @classmethod
    def get_countries(cls) -> List[Country]:
        """Get all countries.

        Returns
        -------
        List[Country]
            List of all countries
        """
        return cls._get_countries().copy()

    @classmethod
    def _get_countries(cls) -> List[Country]:
        """Get all countries from cache.

        The countries list is initialized for the first time and cached
        in internal class attribute.

        Returns
        -------
        List[Country]
            List of all countries
        """
        if cls._COUNTRIES is None:
            cls._COUNTRIES = [Country(row.ISO_3, row.ISO_2, row.NAME)
                              for row in cls._get_countries_df().itertuples()]
        return cls._COUNTRIES

    @classmethod
    def _get_countries_df(cls) -> pandas.DataFrame:
        """Load countries CSV data and save in internal class attribute."""
        if cls._COUNTRIES_DF is None:
            cls._COUNTRIES_DF = pandas.read_csv(tools.get_resource(CountriesDB._COUNTRIES_DB))
        return cls._COUNTRIES_DF



"""The Mimeo CLI Exceptions module.

It contains all custom exceptions related to Mimeo CLI:
    * EnvironmentNotFound
        A custom Exception class for not found environment.
    * EnvironmentsFileNotFound
        A custom Exception class for not found environments file.
"""


class EnvironmentNotFound(Exception):
    """A custom Exception class for not found environment.

    Raised while attempting to access an environment that does not
    exist.
    """

    def __init__(self, env_name: str, envs_file_path: str):
        """Initialize EnvironmentNotFound exception with details.

        Extends Exception constructor with a custom message.

        Parameters
        ----------
        env_name : str
            An environment name
        envs_file_path : str
            An environments file path
        """
        super().__init__(f"No such env [{env_name}] in environments file [{envs_file_path}]")


class EnvironmentsFileNotFound(Exception):
    """A custom Exception class for not found environments file.

    Raised while attempting to access an environments file that does
    not exist.
    """

    def __init__(self, envs_file_path: str):
        """Initialize EnvironmentsFileNotFound exception with details.

        Extends Exception constructor with a custom message.

        Parameters
        ----------
        envs_file_path : str
            An environments file path
        """
        super().__init__(f"Environments file not found [{envs_file_path}]")



"""The Mimeo Context Exceptions module.

It contains all custom exceptions related to Mimeo Context:
    * MinimumIdentifierReached
        A custom Exception class for reaching minimum identifier.
    * UninitializedContextIteration
        A custom Exception class for uninitialized context's iteration.
    * ContextIterationNotFound
        A custom Exception class for not found context's iteration.
    * InvalidSpecialFieldName
        A custom Exception class for invalid special field's name.
    * InvalidSpecialFieldValue
        A custom Exception class for invalid special field's value.
    * SpecialFieldNotFound
        A custom Exception class for not found special field.
    * VarNotFound
        A custom Exception class for not found var.
"""
from typing import Union


class MinimumIdentifierReached(Exception):
    """A custom Exception class for reaching minimum identifier.

    Raised when using MimeoContext.prev_id() method and id is equal to
    0.
    """

    def __init__(self):
        """Initialize MinimumIdentifierReached exception with details.

        Extends Exception constructor with a constant message.
        """
        super().__init__("There's no previous ID!")


class UninitializedContextIteration(Exception):
    """A custom Exception class for uninitialized context's iteration.

    Raised while attempting to access the current iteration without
    prior initialization.
    """

    def __init__(self, context_name: str):
        """Initialize UninitializedContextIteration exception with details.

        Extends Exception constructor with a custom message.

        Parameters
        ----------
        context_name : str
            A current context name
        """
        super().__init__(f"No iteration has been initialized for the current context [{context_name}]")


class ContextIterationNotFound(Exception):
    """A custom Exception class for not found context's iteration.

    Raised while attempting to access an iteration that does not exist.
    """

    def __init__(self, iteration_id: int, context_name: str):
        """Initialize ContextIterationNotFound exception with details.

        Extends Exception constructor with a custom message.

        Parameters
        ----------
        iteration_id : int
            A current context name
        context_name : str
            A current context name
        """
        super().__init__(f"No iteration with id [{iteration_id}] "
                         f"has been initialized for the current context [{context_name}]")


class InvalidSpecialFieldName(Exception):
    """A custom Exception class for invalid special field's name.

    Raised while attempting to save a special field and provided name
    is not a string value.
    """

    def __init__(self):
        """Initialize InvalidSpecialFieldName exception with details.

        Extends Exception constructor with a constant message.
        """
        super().__init__("A special field name needs to be a string value!")


class InvalidSpecialFieldValue(Exception):
    """A custom Exception class for invalid special field's value.

    Raised while attempting to save a special field and provided value
    is non-atomic one.
    """

    def __init__(self, field_value: Union[dict, list]):
        """Initialize InvalidSpecialFieldValue exception with details.

        Extends Exception constructor with a custom message.

        Parameters
        ----------
        field_value : Union[str, int, bool]
            A special field value
        """
        super().__init__(f"Provided field value [{field_value}] is invalid (use any atomic value)!")


class SpecialFieldNotFound(Exception):
    """A custom Exception class for not found special field.

    Raised while attempting to access a special field that does not
    exist.
    """

    def __init__(self, field_name: str):
        """Initialize SpecialFieldNotFound exception with details.

        Extends Exception constructor with a custom message.

        Parameters
        ----------
        field_name : str
            A special field name
        """
        super().__init__(f"Special Field [{field_name}] has not been found!")


class VarNotFound(Exception):
    """A custom Exception class for not found var.

    Raised while attempting to access a variable that does not exist.
    """

    def __init__(self, variable_name: str):
        """Initialize VarNotFound exception with details.

        Extends Exception constructor with a custom message.

        Parameters
        ----------
        variable_name : str
            A variable name
        """
        super().__init__(f"Provided variable [{variable_name}] is not defined!")



"""The Mimeo Iteration module.

It exports only one class:
    * MimeoIteration
        A class representing a single iteration in a Mimeo Template.
"""
import uuid
from typing import Union

from mimeo.context.exc import (InvalidSpecialFieldName,
                               InvalidSpecialFieldValue, SpecialFieldNotFound)


class MimeoIteration:
    """A class representing a single iteration in a Mimeo Template.

    Each iteration has its own id (an ordinal number in a context),
    a key being a unique value across all iterations, and
    it stores special fields that could be ingested in other fields.

    Attributes
    ----------
    id : int
        An ordinal number in a Mimeo Context
    key : str
        An UUID value unique across all templates

    Methods
    -------
    add_special_field(self, field_name: str, field_value: Union[str, int, bool])
        Put a special field entry to memory.
    get_special_field(self, field_name: str) -> Union[str, int, bool]
        Get a special field value from memory.
    """

    def __init__(self, identifier: int):
        """Initialize MimeoIteration class.

        Parameters
        ----------
        identifier : int
            An ordinal number in a Mimeo Context
        """
        self.id = identifier
        self.key = str(uuid.uuid4())
        self._special_fields = {}

    def add_special_field(self, field_name: str, field_value: Union[str, int, bool]):
        """Put a special field entry to memory.

        Parameters
        ----------
        field_name : str
            A special field name
        field_value : Union[str, int, bool]
            A special field value

        Raises
        ------
        InvalidSpecialFieldName
            If the special field name is not a string
        InvalidSpecialFieldValue
            If the special field value is dict or list
        """
        if not isinstance(field_name, str):
            raise InvalidSpecialFieldName()
        if isinstance(field_value, dict) or isinstance(field_value, list):
            raise InvalidSpecialFieldValue(field_value)

        self._special_fields[field_name] = field_value

    def get_special_field(self, field_name: str) -> Union[str, int, bool]:
        """Get a special field value from memory.

        Parameters
        ----------
        field_name : str
            A special field name

        Returns
        -------
        Union[str, int, bool]
            A special field value

        Raises
        ------
        SpecialFieldNotFound
            If the special field does not exist.
        """
        if field_name not in self._special_fields:
            raise SpecialFieldNotFound(field_name)

        return self._special_fields.get(field_name)



"""The Mimeo Generator module.

It exports only one class:
    * Generator
        An abstract class for data generators in Mimeo.
"""
from __future__ import annotations

from abc import ABCMeta, abstractmethod
from typing import Any, Iterator, Union

from mimeo.config.mimeo_config import MimeoConfig, MimeoTemplate


class Generator(metaclass=ABCMeta):
    """An abstract class for data generators in Mimeo.

    Its subclasses are meant to be used in the Mimeo processing.
    Every supported output format has a Generator representation.

    Methods
    -------
    generate(
        templates: Union[list, Iterator[MimeoTemplate]],
        parent: Any = None
    ) -> Iterator[Any]
        Generate data based on the Mimeo Configuration.
    stringify(
        data: Any,
        mimeo_config: MimeoConfig
    ) -> str
        Stringify data generated by the generate() method.
    """

    @classmethod
    def __subclasshook__(
            cls,
            subclass: Generator
    ):
        """Verify if a subclass implements all abstract methods.

        Parameters
        ----------
        subclass : Generator
            A Generator subclass

        Returns
        -------
        bool
            True if the subclass includes the generate and stringify
            methods
        """
        return ('generate' in subclass.__dict__ and callable(subclass.generate) and
                'stringify' in subclass.__dict__ and callable(subclass.stringify))

    @abstractmethod
    def generate(
            self,
            templates: Union[list, Iterator[MimeoTemplate]],
            parent: Any = None
    ) -> Iterator[Any]:
        """Generate data based on the Mimeo Configuration.

        It is an abstract method to implement in subclasses

        Parameters
        ----------
        templates : Union[list, Iterator[MimeoTemplate]]
            A collection of Mimeo Templates to process
        parent : Any, default None
            A parent node for the currently processed template.
            It is passed only when a Mimeo Config contain nested
            templates.

        Returns
        -------
        Iterator[Any]
            Iterator for generated nodes
        """
        pass

    @abstractmethod
    def stringify(
            self,
            data_unit: Any,
            mimeo_config: MimeoConfig
    ) -> str:
        """Stringify data generated by the generate() method.

        It is an abstract method to implement in subclasses

        Parameters
        ----------
        data_unit: Any
            A single data unit generated by the generate() method
        mimeo_config: MimeoConfig
            A Mimeo Configuration providing output details

        Returns
        -------
        str
            Stringified data unit
        """
        pass



"""The Mimeo Configuration module.

It contains classes representing Mimeo Configuration components
at all levels. All of them are Data Transfer Objects:
    * MimeoDTO
        A superclass for all Mimeo configuration DTOs
    * MimeoConfig
        A MimeoDTO class representing Mimeo Configuration
    * MimeoOutputDetails
        A MimeoDTO class representing Mimeo Output Details
    * MimeoTemplate
        A MimeoDTO class representing Mimeo Template
    * MimeoModel
        A MimeoDTO class representing Mimeo Model
"""
import re

from mimeo.config.exc import (InvalidIndent, InvalidMimeoConfig,
                              InvalidMimeoModel, InvalidMimeoTemplate,
                              InvalidVars, MissingRequiredProperty,
                              UnsupportedPropertyValue)
from mimeo.logging import setup_logging

# setup logging when mimeo is used as a python library
setup_logging()


class MimeoDTO:
    """A superclass for all Mimeo configuration DTOs.

    It is meant to store a source dictionary for logging purposes.

    Methods
    -------
    __str__
        Return the stringified source dictionary of a DTO.
    """

    def __init__(self, source: dict):
        """Initialize MimeoDTO class.

        Parameters
        ----------
        source : dict
            The source dictionary for a Mimeo DTO
        """
        self._source = source

    def __str__(self):
        """Return the stringified source dictionary of a DTO."""
        return str(self._source)


class MimeoConfig(MimeoDTO):
    """A MimeoDTO class representing Mimeo Configuration.

    It is a python representation of a Mimeo Configuration file / dictionary.

    Attributes
    ----------
    OUTPUT_DETAILS_KEY : str
        A Mimeo Configuration output details key
    OUTPUT_DETAILS_DIRECTION_KEY : str
        A Mimeo Configuration output direction key
    OUTPUT_DETAILS_FORMAT_KEY : str
        A Mimeo Configuration output format key
    OUTPUT_DETAILS_XML_DECLARATION_KEY : str
        A Mimeo Configuration xml declaration key
    OUTPUT_DETAILS_INDENT_KEY : str
        A Mimeo Configuration indent key
    OUTPUT_DETAILS_DIRECTORY_PATH_KEY : str
        A Mimeo Configuration output directory path key
    OUTPUT_DETAILS_FILE_NAME_KEY : str
        A Mimeo Configuration output file name key
    OUTPUT_DETAILS_METHOD_KEY : str
        A Mimeo Configuration http method key
    OUTPUT_DETAILS_PROTOCOL_KEY : str
        A Mimeo Configuration http protocol key
    OUTPUT_DETAILS_HOST_KEY : str
        A Mimeo Configuration http host key
    OUTPUT_DETAILS_PORT_KEY : str
        A Mimeo Configuration http port key
    OUTPUT_DETAILS_ENDPOINT_KEY : str
        A Mimeo Configuration http endpoint key
    OUTPUT_DETAILS_AUTH_KEY : str
        A Mimeo Configuration http auth key
    OUTPUT_DETAILS_USERNAME_KEY : str
        A Mimeo Configuration http username key
    OUTPUT_DETAILS_PASSWORD_KEY : str
        A Mimeo Configuration http password key
    VARS_KEY : str
        A Mimeo Configuration vars key
    TEMPLATES_KEY : str
        A Mimeo Configuration templates key
    TEMPLATES_COUNT_KEY : str
        A Mimeo Configuration template's count key
    TEMPLATES_MODEL_KEY : str
        A Mimeo Configuration template's model key
    MODEL_CONTEXT_KEY : str
        A Mimeo Configuration model's context name key
    MODEL_ATTRIBUTES_KEY : str
        A Mimeo Configuration attributes key (for nodes' attributes)
    MODEL_VALUE_KEY : str
        A Mimeo Configuration value key (for nodes' value)
    MODEL_MIMEO_UTIL_KEY : str
        A Mimeo Configuration Mimeo Util key
    MODEL_MIMEO_UTIL_NAME_KEY : str
        A Mimeo Configuration Mimeo Util's name key
    SUPPORTED_OUTPUT_FORMATS : set
        A set of supported output formats
    OUTPUT_DETAILS_DIRECTION_FILE : str
        The 'file' output direction
    OUTPUT_DETAILS_DIRECTION_STD_OUT : str
        The 'stdout' output direction
    OUTPUT_DETAILS_DIRECTION_HTTP : str
        The 'http' output direction
    OUTPUT_DETAILS_DIRECTION_HTTP_REQUEST_POST : str
        The 'POST' http request method
    OUTPUT_DETAILS_DIRECTION_HTTP_REQUEST_PUT : str
        The 'PUT' http request method
    OUTPUT_DETAILS_DIRECTION_HTTP_AUTH_BASIC : str
        The 'basic' http auth method
    OUTPUT_DETAILS_DIRECTION_HTTP_AUTH_DIGEST : str
        The 'digest' http auth method
    SUPPORTED_OUTPUT_DIRECTIONS : set
        List of supported output directions
    SUPPORTED_REQUEST_METHODS : set
        List of supported http request methods
    SUPPORTED_AUTH_METHODS : set
        List of supported auth request methods
    REQUIRED_HTTP_DETAILS : set
        List of required http request output direction details

    output_details : MimeoOutputDetails, default {}
        A Mimeo Output Details settings
    vars : dict, default {}
        A Mimeo Configuration vars setting
    templates : list
        A Mimeo Templates setting
    """

    OUTPUT_DETAILS_KEY = "output_details"
    OUTPUT_DETAILS_DIRECTION_KEY = "direction"
    OUTPUT_DETAILS_FORMAT_KEY = "format"
    OUTPUT_DETAILS_XML_DECLARATION_KEY = "xml_declaration"
    OUTPUT_DETAILS_INDENT_KEY = "indent"
    OUTPUT_DETAILS_DIRECTORY_PATH_KEY = "directory_path"
    OUTPUT_DETAILS_FILE_NAME_KEY = "file_name"
    OUTPUT_DETAILS_METHOD_KEY = "method"
    OUTPUT_DETAILS_PROTOCOL_KEY = "protocol"
    OUTPUT_DETAILS_HOST_KEY = "host"
    OUTPUT_DETAILS_PORT_KEY = "port"
    OUTPUT_DETAILS_ENDPOINT_KEY = "endpoint"
    OUTPUT_DETAILS_AUTH_KEY = "auth"
    OUTPUT_DETAILS_USERNAME_KEY = "username"
    OUTPUT_DETAILS_PASSWORD_KEY = "password"
    VARS_KEY = "vars"
    TEMPLATES_KEY = "_templates_"
    TEMPLATES_COUNT_KEY = "count"
    TEMPLATES_MODEL_KEY = "model"
    MODEL_CONTEXT_KEY = "context"
    MODEL_ATTRIBUTES_KEY = "_attrs"
    MODEL_VALUE_KEY = "_value"
    MODEL_MIMEO_UTIL_KEY = "_mimeo_util"
    MODEL_MIMEO_UTIL_NAME_KEY = "_name"

    OUTPUT_FORMAT_XML = "xml"

    OUTPUT_DETAILS_DIRECTION_FILE = "file"
    OUTPUT_DETAILS_DIRECTION_STD_OUT = "stdout"
    OUTPUT_DETAILS_DIRECTION_HTTP = "http"

    OUTPUT_DETAILS_DIRECTION_HTTP_REQUEST_POST = "POST"
    OUTPUT_DETAILS_DIRECTION_HTTP_REQUEST_PUT = "PUT"

    OUTPUT_DETAILS_DIRECTION_HTTP_AUTH_BASIC = "basic"
    OUTPUT_DETAILS_DIRECTION_HTTP_AUTH_DIGEST = "digest"

    SUPPORTED_OUTPUT_FORMATS = (OUTPUT_FORMAT_XML,)

    SUPPORTED_OUTPUT_DIRECTIONS = (OUTPUT_DETAILS_DIRECTION_STD_OUT,
                                   OUTPUT_DETAILS_DIRECTION_FILE,
                                   OUTPUT_DETAILS_DIRECTION_HTTP)
    SUPPORTED_REQUEST_METHODS = (OUTPUT_DETAILS_DIRECTION_HTTP_REQUEST_POST,
                                 OUTPUT_DETAILS_DIRECTION_HTTP_REQUEST_PUT)
    SUPPORTED_AUTH_METHODS = (OUTPUT_DETAILS_DIRECTION_HTTP_AUTH_BASIC,
                              OUTPUT_DETAILS_DIRECTION_HTTP_AUTH_DIGEST)
    REQUIRED_HTTP_DETAILS = (OUTPUT_DETAILS_HOST_KEY,
                             OUTPUT_DETAILS_ENDPOINT_KEY,
                             OUTPUT_DETAILS_USERNAME_KEY,
                             OUTPUT_DETAILS_PASSWORD_KEY)

    def __init__(self, config: dict):
        """Initialize MimeoConfig class.

        Extends MimeoDTO constructor.

        Parameters
        ----------
        config : dict
            A source config dictionary
        """
        super().__init__(config)
        self.output_details = MimeoOutputDetails(config.get(self.OUTPUT_DETAILS_KEY, {}))
        self.vars = MimeoConfig._get_vars(config)
        self.templates = MimeoConfig._get_templates(config)

    @staticmethod
    def _get_vars(config: dict) -> dict:
        """Extract variables from the source dictionary.

        Parameters
        ----------
        config : dict
            A source config dictionary

        Returns
        -------
        variables : dict
            Customized variables or an empty dictionary

        Raises
        ------
        InvalidVars
            If (1) the vars key does not point to a dictionary or
            (2) some variable's name does not start with a letter,
            is not SNAKE_UPPER_CASE with possible digits or
            (3) some variable's value points to non-atomic value nor Mimeo Util
        """
        variables = config.get(MimeoConfig.VARS_KEY, {})
        if not isinstance(variables, dict):
            raise InvalidVars(f"vars property does not store an object: {variables}")
        for var, val in variables.items():
            if not re.match(r"^[A-Z][A-Z_0-9]*$", var):
                raise InvalidVars(f"Provided var [{var}] is invalid "
                                  "(you can use upper-cased name with underscore and digits, starting with a letter)!")
            if isinstance(val, list) or (isinstance(val, dict) and not MimeoConfig._is_mimeo_util_object(val)):
                raise InvalidVars(f"Provided var [{var}] is invalid (you can use ony atomic values and Mimeo Utils)!")
        return variables

    @staticmethod
    def _get_templates(config: dict) -> list:
        """Extract Mimeo Templates from the source dictionary.

        Parameters
        ----------
        config : dict
            A source config dictionary

        Returns
        -------
        list
            A Mimeo Templates list

        Raises
        ------
        IncorrectMimeoConfig
            If (1) the source dictionary does not include the _templates_ key or
            (2) the _templates_ key does not point to a list
        """
        templates = config.get(MimeoConfig.TEMPLATES_KEY)
        if templates is None:
            raise InvalidMimeoConfig(f"No templates in the Mimeo Config: {config}")
        elif not isinstance(templates, list):
            raise InvalidMimeoConfig(f"_templates_ property does not store an array: {config}")
        else:
            return [MimeoTemplate(template) for template in config.get(MimeoConfig.TEMPLATES_KEY)]

    @staticmethod
    def _is_mimeo_util_object(obj: dict) -> bool:
        """Verify if the object is a Mimeo Util.

        Parameters
        ----------
        obj : dict
            An object to verify

        Returns
        -------
        bool
            True if the object is a dictionary having only one key: _mimeo_util, otherwise False
        """
        return isinstance(obj, dict) and len(obj) == 1 and MimeoConfig.MODEL_MIMEO_UTIL_KEY in obj


class MimeoOutputDetails(MimeoDTO):
    """A MimeoDTO class representing Mimeo Output Details.

    It is a python representation of a Mimeo Output Details configuration node.

    Attributes
    ----------
    direction : str, default 'file'
        The configured output direction
    format : str, default 'xml'
        A Mimeo Configuration output format setting
    xml_declaration : bool, default False
        A Mimeo Configuration xml declaration setting
    indent : int, default 0
        A Mimeo Configuration indent setting
    directory_path : str, default 'mimeo-output'
        The configured file output directory
    file_name_tmplt : str, default 'mimeo-output-{}.{output_format}'
        The configured file output file name template
    method : str, default POST
        The configured http output request method
    protocol : str, default 'http'
        The configured http output protocol
    host : str
        The configured http output host
    port : str
        The configured http output port
    endpoint : str
        The configured http output endpoint
    auth : str, default 'basic'
        The configured http output auth method
    username : str
        The configured http output username
    password : str
        The configured http output password
    """

    def __init__(self, output_details: dict):
        """Initialize MimeoOutputDetails class.

        Extends MimeoDTO constructor.

        Parameters
        ----------
        output_details : dict
            A source config output details dictionary
        """
        super().__init__(output_details)
        self.direction = MimeoOutputDetails._get_direction(output_details)
        MimeoOutputDetails._validate_output_details(self.direction, output_details)
        self.format = MimeoOutputDetails._get_format(output_details)
        self.xml_declaration = output_details.get(MimeoConfig.OUTPUT_DETAILS_XML_DECLARATION_KEY, False)
        self.indent = MimeoOutputDetails._get_indent(output_details)
        self.directory_path = MimeoOutputDetails._get_directory_path(self.direction, output_details)
        self.file_name_tmplt = MimeoOutputDetails._get_file_name_tmplt(self.direction, output_details, self.format)
        self.method = MimeoOutputDetails._get_method(self.direction, output_details)
        self.protocol = MimeoOutputDetails._get_protocol(self.direction, output_details)
        self.host = MimeoOutputDetails._get_host(self.direction, output_details)
        self.port = MimeoOutputDetails._get_port(self.direction, output_details)
        self.endpoint = MimeoOutputDetails._get_endpoint(self.direction, output_details)
        self.auth = MimeoOutputDetails._get_auth(self.direction, output_details)
        self.username = MimeoOutputDetails._get_username(self.direction, output_details)
        self.password = MimeoOutputDetails._get_password(self.direction, output_details)

    @staticmethod
    def _get_direction(output_details: dict) -> str:
        """Extract output direction from the source dictionary.

        Parameters
        ----------
        output_details : dict
            A source config output details dictionary

        Returns
        -------
        direction : str
            The configured output direction

        Raises
        ------
        UnsupportedPropertyValue
            If the configured output direction is not supported
        """
        direction = output_details.get(MimeoConfig.OUTPUT_DETAILS_DIRECTION_KEY,
                                       MimeoConfig.OUTPUT_DETAILS_DIRECTION_FILE)
        if direction in MimeoConfig.SUPPORTED_OUTPUT_DIRECTIONS:
            return direction
        else:
            raise UnsupportedPropertyValue(MimeoConfig.OUTPUT_DETAILS_DIRECTION_KEY,
                                           direction,
                                           MimeoConfig.SUPPORTED_OUTPUT_DIRECTIONS)

    @staticmethod
    def _get_format(config: dict) -> str:
        """Extract an output format from the source dictionary.

        Parameters
        ----------
        config : dict
            A source config dictionary

        Returns
        -------
        output_format : str
            The customized output format or 'xml' by default

        Raises
        ------
        UnsupportedPropertyValue
            If the customized output format is not supported
        """
        output_format = config.get(MimeoConfig.OUTPUT_DETAILS_FORMAT_KEY, MimeoConfig.OUTPUT_FORMAT_XML)
        if output_format in MimeoConfig.SUPPORTED_OUTPUT_FORMATS:
            return output_format
        else:
            raise UnsupportedPropertyValue(MimeoConfig.OUTPUT_DETAILS_FORMAT_KEY,
                                           output_format,
                                           MimeoConfig.SUPPORTED_OUTPUT_FORMATS)

    @staticmethod
    def _get_indent(config: dict) -> int:
        """Extract an indent value from the source dictionary.

        Parameters
        ----------
        config : dict
            A source config dictionary

        Returns
        -------
        indent : int
            The customized indent or 0 by default

        Raises
        ------
        InvalidIndent
            If the customized indent is lower than zero
        """
        indent = config.get(MimeoConfig.OUTPUT_DETAILS_INDENT_KEY, 0)
        if indent >= 0:
            return indent
        else:
            raise InvalidIndent(indent)

    @staticmethod
    def _get_directory_path(direction: str, output_details: dict) -> str:
        """Extract an output directory path from the source dictionary.

        It is extracted only when the output direction is 'file'.

        Parameters
        ----------
        direction : str
            The configured output direction
        output_details : dict
            A source config output details dictionary

        Returns
        -------
        str
            The configured output directory path when the output direction is 'file'.
            Otherwise, None. If the 'directory_path' setting is missing returns
            'mimeo-output' by default.
        """
        if direction == MimeoConfig.OUTPUT_DETAILS_DIRECTION_FILE:
            return output_details.get(MimeoConfig.OUTPUT_DETAILS_DIRECTORY_PATH_KEY, "mimeo-output")

    @staticmethod
    def _get_file_name_tmplt(direction: str, output_details: dict, output_format: str):
        """Generate an output file name template based on the source dictionary.

        It is generated only when the output direction is 'file'.

        Parameters
        ----------
        direction : str
            The configured output direction
        output_details : dict
            A source config output details dictionary

        Returns
        -------
        str
            The configured output file name template when the output direction is 'file'.
            Otherwise, None. If the 'file_name' setting is missing returns
            'mimeo-output-{}.{output_format}' by default.
        """
        if direction == MimeoConfig.OUTPUT_DETAILS_DIRECTION_FILE:
            file_name = output_details.get(MimeoConfig.OUTPUT_DETAILS_FILE_NAME_KEY, "mimeo-output")
            return f"{file_name}-{'{}'}.{output_format}"

    @staticmethod
    def _get_method(direction: str, output_details: dict) -> str:
        """Extract an HTTP request method from the source dictionary.

        It is extracted only when the output direction is 'http'.

        Parameters
        ----------
        direction : str
            The configured output direction
        output_details : dict
            A source config output details dictionary

        Returns
        -------
        method: str
            The configured HTTP request method when the output direction is 'http'.
            Otherwise, None. If the 'method' setting is missing returns
            'POST' by default.
        """
        if direction == MimeoConfig.OUTPUT_DETAILS_DIRECTION_HTTP:
            method = output_details.get(MimeoConfig.OUTPUT_DETAILS_METHOD_KEY,
                                        MimeoConfig.OUTPUT_DETAILS_DIRECTION_HTTP_REQUEST_POST)
            if method in MimeoConfig.SUPPORTED_REQUEST_METHODS:
                return method
            else:
                raise UnsupportedPropertyValue(MimeoConfig.OUTPUT_DETAILS_METHOD_KEY,
                                               method,
                                               MimeoConfig.SUPPORTED_REQUEST_METHODS)

    @staticmethod
    def _get_protocol(direction: str, output_details: dict) -> str:
        """Extract an HTTP protocol from the source dictionary.

        It is extracted only when the output direction is 'http'.

        Parameters
        ----------
        direction : str
            The configured output direction
        output_details : dict
            A source config output details dictionary

        Returns
        -------
        str
            The configured HTTP request method when the output direction is 'http'.
            Otherwise, None. If the 'protocol' setting is missing returns
            'http' by default.
        """
        if direction == MimeoConfig.OUTPUT_DETAILS_DIRECTION_HTTP:
            return output_details.get(MimeoConfig.OUTPUT_DETAILS_PROTOCOL_KEY,
                                      MimeoConfig.OUTPUT_DETAILS_DIRECTION_HTTP)

    @staticmethod
    def _get_host(direction: str, output_details: dict) -> str:
        """Extract an HTTP host from the source dictionary.

        It is extracted only when the output direction is 'http'.

        Parameters
        ----------
        direction : str
            The configured output direction
        output_details : dict
            A source config output details dictionary

        Returns
        -------
        str
            The configured HTTP host when the output direction is 'http'.
            Otherwise, None.
        """
        if direction == MimeoConfig.OUTPUT_DETAILS_DIRECTION_HTTP:
            return output_details.get(MimeoConfig.OUTPUT_DETAILS_HOST_KEY)

    @staticmethod
    def _get_port(direction: str, output_details: dict) -> str:
        """Extract an HTTP port from the source dictionary.

        It is extracted only when the output direction is 'http'.

        Parameters
        ----------
        direction : str
            The configured output direction
        output_details : dict
            A source config output details dictionary

        Returns
        -------
        str
            The configured HTTP port when the output direction is 'http'.
            Otherwise, None.
        """
        if direction == MimeoConfig.OUTPUT_DETAILS_DIRECTION_HTTP:
            return output_details.get(MimeoConfig.OUTPUT_DETAILS_PORT_KEY)

    @staticmethod
    def _get_endpoint(direction: str, output_details: dict) -> str:
        """Extract an HTTP endpoint from the source dictionary.

        It is extracted only when the output direction is 'http'.

        Parameters
        ----------
        direction : str
            The configured output direction
        output_details : dict
            A source config output details dictionary

        Returns
        -------
        str
            The configured HTTP request method when the output direction is 'http'.
            Otherwise, None.
        """
        if direction == MimeoConfig.OUTPUT_DETAILS_DIRECTION_HTTP:
            return output_details.get(MimeoConfig.OUTPUT_DETAILS_ENDPOINT_KEY)

    @staticmethod
    def _get_auth(direction: str, output_details: dict) -> str:
        """Extract an HTTP auth method from the source dictionary.

        It is extracted only when the output direction is 'http'.

        Parameters
        ----------
        direction : str
            The configured output direction
        output_details : dict
            A source config output details dictionary

        Returns
        -------
        method: str
            The configured HTTP auth method when the output direction is 'http'.
            Otherwise, None. If the 'auth' setting is missing returns
            'basic' by default.
        """
        if direction == MimeoConfig.OUTPUT_DETAILS_DIRECTION_HTTP:
            auth =  output_details.get(MimeoConfig.OUTPUT_DETAILS_AUTH_KEY,
                                       MimeoConfig.OUTPUT_DETAILS_DIRECTION_HTTP_AUTH_BASIC)
            if auth in MimeoConfig.SUPPORTED_AUTH_METHODS:
                return auth
            else:
                raise UnsupportedPropertyValue(MimeoConfig.OUTPUT_DETAILS_AUTH_KEY,
                                               auth,
                                               MimeoConfig.SUPPORTED_AUTH_METHODS)

    @staticmethod
    def _get_username(direction: str, output_details: dict) -> str:
        """Extract a username from the source dictionary.

        It is extracted only when the output direction is 'http'.

        Parameters
        ----------
        direction : str
            The configured output direction
        output_details : dict
            A source config output details dictionary

        Returns
        -------
        str
            The configured username when the output direction is 'http'.
            Otherwise, None.
        """
        if direction == MimeoConfig.OUTPUT_DETAILS_DIRECTION_HTTP:
            return output_details.get(MimeoConfig.OUTPUT_DETAILS_USERNAME_KEY)

    @staticmethod
    def _get_password(direction: str, output_details: dict) -> str:
        """Extract a password from the source dictionary.

        It is extracted only when the output direction is 'http'.

        Parameters
        ----------
        direction : str
            The configured output direction
        output_details : dict
            A source config output details dictionary

        Returns
        -------
        str
            The configured password when the output direction is 'http'.
            Otherwise, None.
        """
        if direction == MimeoConfig.OUTPUT_DETAILS_DIRECTION_HTTP:
            return output_details.get(MimeoConfig.OUTPUT_DETAILS_PASSWORD_KEY)

    @staticmethod
    def _validate_output_details(direction: str, output_details: dict) -> None:
        """Validate output details in the source dictionary.

        The validation is being done according to the configured output
        direction.

        Parameters
        ----------
        direction : str
            The configured output direction
        output_details : dict
            A source config output details dictionary

        Raises
        ------
        MissingRequiredProperty
            If the output details doesn't include all required settings for the direction
        """
        if direction == MimeoConfig.OUTPUT_DETAILS_DIRECTION_HTTP:
            missing_details = []
            for detail in MimeoConfig.REQUIRED_HTTP_DETAILS:
                if detail not in output_details:
                    missing_details.append(detail)
            if len(missing_details) > 0:
                missing_details_str = ', '.join(missing_details)
                raise MissingRequiredProperty(f"Missing required fields is HTTP output details: {missing_details_str}")


class MimeoTemplate(MimeoDTO):
    """A MimeoDTO class representing Mimeo Template.

    It is a python representation of a Mimeo Template configuration node.

    Attributes
    ----------
    count : int
        A configured count of the Mimeo Template
    model : MimeoModel
        A configured model of the Mimeo Template
    """

    def __init__(self, template: dict):
        """Initialize MimeoTemplate class.

        Extends MimeoDTO constructor.

        Parameters
        ----------
        template : dict
            A source config template dictionary
        """
        super().__init__(template)
        MimeoTemplate._validate_template(template)
        self.count = template.get(MimeoConfig.TEMPLATES_COUNT_KEY)
        self.model = MimeoModel(template.get(MimeoConfig.TEMPLATES_MODEL_KEY))

    @staticmethod
    def _validate_template(template: dict) -> None:
        """Validate template in the source dictionary.

        Parameters
        ----------
        template : dict
            A source config template dictionary

        Raises
        ------
        IncorrectMimeoTemplate
            If the source config doesn't include count or model properties
        """
        if MimeoConfig.TEMPLATES_COUNT_KEY not in template:
            raise InvalidMimeoTemplate(f"No count value in the Mimeo Template: {template}")
        if MimeoConfig.TEMPLATES_MODEL_KEY not in template:
            raise InvalidMimeoTemplate(f"No model data in the Mimeo Template: {template}")


class MimeoModel(MimeoDTO):
    """A MimeoDTO class representing Mimeo Model.

    It is a python representation of a Mimeo Model configuration node.

    Attributes
    ----------
    root_name : str
        A root node's tag
    root_data : dict
        A template data
    context_name : str
        A context name (root_name by default)
    """

    def __init__(self, model: dict):
        """Initialize MimeoModel class.

        Extends MimeoDTO constructor.

        Parameters
        ----------
        model : dict
            A source config model dictionary
        """
        super().__init__(model)
        self.root_name = MimeoModel._get_root_name(model)
        self.root_data = model.get(self.root_name)
        self.context_name = MimeoModel._get_context_name(model, self.root_name)

    @staticmethod
    def _get_root_name(model: dict) -> str:
        """Extract a root name from the source dictionary.

        Parameters
        ----------
        model : dict
            A source config model dictionary

        Returns
        -------
        str
            The configured root node's tag

        Raises
        ------
        IncorrectMimeoModel
            If the source config has no or more than one root nodes
        """
        model_keys = [key for key in filter(MimeoModel._is_not_configuration_key, iter(model))]
        if len(model_keys) == 1:
            return model_keys[0]
        if len(model_keys) == 0:
            raise InvalidMimeoModel(f"No root data in Mimeo Model: {model}")
        elif len(model_keys) > 1:
            raise InvalidMimeoModel(f"Multiple root data in Mimeo Model: {model}")

    @staticmethod
    def _get_context_name(model: dict, root_name: str) -> str:
        """Extract a context name from the source dictionary.

        Parameters
        ----------
        model : dict
            A source config model dictionary
        root_name : str
            The configured root node's tag

        Returns
        -------
        str
            The configured context name.
            If the 'context' setting is missing returns root name by default

        Raises
        ------
        IncorrectMimeoModel
            If the source config has a context name not being a string value
        """
        context_name = model.get(MimeoConfig.MODEL_CONTEXT_KEY, root_name)
        if isinstance(context_name, str):
            return context_name
        else:
            raise InvalidMimeoModel(f"Invalid context name in Mimeo Model (not a string value): {model}")

    @staticmethod
    def _is_not_configuration_key(dict_key: str) -> bool:
        """Verify if the dictionary key is a configuration one.

        Parameters
        ----------
        dict_key : str
            A dictionary key to verify

        Returns
        -------
        bool
            True if the key is 'context', otherwise False
        """
        return dict_key not in [MimeoConfig.MODEL_CONTEXT_KEY]



"""The Mimeo Parser module.

It exports two parser classes:
    * MimeoArgumentParser
        A custom ArgumentParser for the Mimeo CLI.
    * MimeoConfigParser
        A class parsing source Mimeo Configuration with Mimeo
        arguments.
"""
import json
import logging
from argparse import ArgumentParser, Namespace
from os import path
from typing import Union

from mimeo import MimeoConfig
from mimeo.cli.exc import EnvironmentNotFound, EnvironmentsFileNotFound

logger = logging.getLogger(__name__)

DEFAULT_ENVS_FILE_PATH = "mimeo.envs.json"


class MimeoArgumentParser(ArgumentParser):
    """A custom ArgumentParser for the Mimeo CLI."""

    def __init__(self):
        """Initialize MimeoArgumentParser class.

        Extends ArgumentParser constructor with a Mimeo CLI details.

        It provides the following command line interface:

        usage: mimeo [OPTIONS] paths

        Generate data based on a template

        positional arguments:
          paths                 take paths to Mimeo Configuration files

        optional arguments:
          -h, --help            show this help message and exit
          -v, --version         show program's version number and exit

        Mimeo Configuration arguments:
          -o {file,stdout,http}, --output {file,stdout,http}
                                overwrite the output_details/direction property
          -x {true,false}, --xml-declaration {true,false}
                                overwrite the output_details/xml_declaration property
          -i INDENT, --indent INDENT
                                overwrite the output_details/indent property
          -d DIRECTORY_PATH, --directory DIRECTORY_PATH
                                overwrite the output_details/directory_path property
          -f FILE_NAME, --file FILE_NAME
                                overwrite the output_details/file_name property
          -H HOST, --http-host HOST
                                overwrite the output_details/host property
          -p PORT, --http-port PORT
                                overwrite the output_details/port property
          -E ENDPOINT, --http-endpoint ENDPOINT
                                overwrite the output_details/endpoint property
          -U USERNAME, --http-user USERNAME
                                overwrite the output_details/username property
          -P PASSWORD, --http-password PASSWORD
                                overwrite the output_details/password property
          --http-method METHOD
                                overwrite the output_details/method property
          --http-protocol PROTOCOL
                                overwrite the output_details/protocol property
          --http-auth AUTH
                                overwrite the output_details/auth property
          -e ENVIRONMENT, --http-env ENVIRONMENT
                                overwrite the output_details http properties using a mimeo environment configuration
          --http-envs-file PATH
                                use a custom environments file (by default: mimeo.envs.json)

        Logging arguments:
          --silent              disable INFO logs
          --debug               enable DEBUG mode
          --fine                enable FINE mode
        """
        super().__init__(
            prog="mimeo",
            description="Generate data based on a template",
            usage="%(prog)s [OPTIONS] paths")
        self._add_positional_arguments()
        self._add_mimeo_configuration_arguments()
        self._add_logging_arguments()

    def _add_positional_arguments(self):
        """Add positional arguments."""
        self.add_argument(
            "-v",
            "--version",
            action="version",
            version="%(prog)s v0.4.2")
        self.add_argument(
            "paths",
            nargs="+",
            type=str,
            help="take paths to Mimeo Configuration files")

    def _add_mimeo_configuration_arguments(self):
        """Add arguments overwriting Mimeo Configuration."""
        mimeo_config_args = self.add_argument_group("Mimeo Configuration arguments")
        mimeo_config_args.add_argument(
            "-o",
            "--output",
            type=str,
            choices=["file", "stdout", "http"],
            help="overwrite the output_details/direction property")
        mimeo_config_args.add_argument(
            "-x",
            "--xml-declaration",
            type=str,
            choices=["true", "false"],
            help="overwrite the output_details/xml_declaration property")
        mimeo_config_args.add_argument(
            "-i",
            "--indent",
            type=int,
            help="overwrite the output_details/indent property")
        mimeo_config_args.add_argument(
            "-d",
            "--directory",
            type=str,
            metavar="DIRECTORY_PATH",
            help="overwrite the output_details/directory_path property")
        mimeo_config_args.add_argument(
            "-f",
            "--file",
            type=str,
            metavar="FILE_NAME",
            help="overwrite the output_details/file_name property")
        mimeo_config_args.add_argument(
            "-H",
            "--http-host",
            type=str,
            metavar="HOST",
            help="overwrite the output_details/host property")
        mimeo_config_args.add_argument(
            "-p",
            "--http-port",
            type=str,
            metavar="PORT",
            help="overwrite the output_details/port property")
        mimeo_config_args.add_argument(
            "-E",
            "--http-endpoint",
            type=str,
            metavar="ENDPOINT",
            help="overwrite the output_details/endpoint property")
        mimeo_config_args.add_argument(
            "-U",
            "--http-user",
            type=str,
            metavar="USERNAME",
            help="overwrite the output_details/username property")
        mimeo_config_args.add_argument(
            "-P",
            "--http-password",
            type=str,
            metavar="PASSWORD",
            help="overwrite the output_details/password property")
        mimeo_config_args.add_argument(
            "--http-method",
            type=str,
            metavar="METHOD",
            help="overwrite the output_details/method property")
        mimeo_config_args.add_argument(
            "--http-protocol",
            type=str,
            metavar="PROTOCOL",
            help="overwrite the output_details/protocol property")
        mimeo_config_args.add_argument(
            "--http-auth",
            type=str,
            metavar="AUTH",
            help="overwrite the output_details/auth property")
        mimeo_config_args.add_argument(
            "-e",
            "--http-env",
            type=str,
            metavar="ENVIRONMENT",
            help="overwrite the output_details http properties using a mimeo environment configuration")
        mimeo_config_args.add_argument(
            "--http-envs-file",
            type=str,
            metavar="PATH",
            help=f"use a custom environments file (by default: {DEFAULT_ENVS_FILE_PATH})")

    def _add_logging_arguments(self):
        """Add arguments customizing logs producing."""
        logging_args = self.add_argument_group("Logging arguments")
        logging_args_excl = logging_args.add_mutually_exclusive_group()
        logging_args_excl.add_argument(
            "--silent",
            action="store_true",
            help="disable INFO logs")
        logging_args_excl.add_argument(
            "--debug",
            action="store_true",
            help="enable DEBUG mode")
        logging_args_excl.add_argument(
            "--fine",
            action="store_true",
            help="enable FINE mode")


class MimeoConfigParser:
    """A class parsing source Mimeo Configuration with Mimeo arguments.

    Methods
    -------
    parse_config() -> MimeoConfig
        Parse a Mimeo Configuration using Mimeo arguments.
    """

    _ENVIRONMENT_PROPS = [MimeoConfig.OUTPUT_DETAILS_PROTOCOL_KEY,
                          MimeoConfig.OUTPUT_DETAILS_HOST_KEY,
                          MimeoConfig.OUTPUT_DETAILS_PORT_KEY,
                          MimeoConfig.OUTPUT_DETAILS_AUTH_KEY,
                          MimeoConfig.OUTPUT_DETAILS_USERNAME_KEY,
                          MimeoConfig.OUTPUT_DETAILS_PASSWORD_KEY]
    
    _ENTRY_PATH_KEY = "entry_path"
    _GET_VALUE_KEY = "get_value"
    _ARGS_MAPPING = {
        "output": {
            "entry_path": [MimeoConfig.OUTPUT_DETAILS_KEY, MimeoConfig.OUTPUT_DETAILS_DIRECTION_KEY]
        },
        "xml_declaration": {
            "entry_path": [MimeoConfig.OUTPUT_DETAILS_KEY, MimeoConfig.OUTPUT_DETAILS_XML_DECLARATION_KEY],
            "get_value": lambda arg: arg.lower() == "true"
        },
        "indent": {
            "entry_path": [MimeoConfig.OUTPUT_DETAILS_KEY, MimeoConfig.OUTPUT_DETAILS_INDENT_KEY]
        },
        "directory": {
            "entry_path": [MimeoConfig.OUTPUT_DETAILS_KEY, MimeoConfig.OUTPUT_DETAILS_DIRECTORY_PATH_KEY]
        },
        "file": {
            "entry_path": [MimeoConfig.OUTPUT_DETAILS_KEY, MimeoConfig.OUTPUT_DETAILS_FILE_NAME_KEY]
        },
        "http_method": {
            "entry_path": [MimeoConfig.OUTPUT_DETAILS_KEY, MimeoConfig.OUTPUT_DETAILS_METHOD_KEY]
        },
        "http_protocol": {
            "entry_path": [MimeoConfig.OUTPUT_DETAILS_KEY, MimeoConfig.OUTPUT_DETAILS_PROTOCOL_KEY]
        },
        "http_auth": {
            "entry_path": [MimeoConfig.OUTPUT_DETAILS_KEY, MimeoConfig.OUTPUT_DETAILS_AUTH_KEY]
        },
        "http_host": {
            "entry_path": [MimeoConfig.OUTPUT_DETAILS_KEY, MimeoConfig.OUTPUT_DETAILS_HOST_KEY]
        },
        "http_port": {
            "entry_path": [MimeoConfig.OUTPUT_DETAILS_KEY, MimeoConfig.OUTPUT_DETAILS_PORT_KEY]
        },
        "http_endpoint": {
            "entry_path": [MimeoConfig.OUTPUT_DETAILS_KEY, MimeoConfig.OUTPUT_DETAILS_ENDPOINT_KEY]
        },
        "http_user": {
            "entry_path": [MimeoConfig.OUTPUT_DETAILS_KEY, MimeoConfig.OUTPUT_DETAILS_USERNAME_KEY]
        },
        "http_password": {
            "entry_path": [MimeoConfig.OUTPUT_DETAILS_KEY, MimeoConfig.OUTPUT_DETAILS_PASSWORD_KEY]
        }
    }

    def __init__(self, config: dict, args: Namespace):
        """Initialize MimeoConfigParser class.

        Parameters
        ----------
        config:
            A source config dictionary
        args
            Arguments parsed by MimeoArgumentParser
        """
        self._raw_config = config
        self._args = args

    def parse_config(self) -> MimeoConfig:
        """Parse a Mimeo Configuration using Mimeo arguments.

        The parsing is made according to the following rule:
        configuration is overwritten from the most general to the most
        specific setting. It means that, whenever CLI arguments
        are overlapping (e.g. -e and -u), first the higher-level
        customization is applied (e.g. environment) and then the lower
        one (e.g. username).

        Returns
        -------
        mimeo_config : MimeoConfig
            A parsed Mimeo Configuration

        Raises
        ------
        EnvironmentsFileNotFound
            If environments file does not exist.
        EnvironmentNotFound
            If the http environment is not defined in the environments file
        """
        logger.info("Parsing Mimeo Configuration")
        config = dict(self._raw_config)
        config = self._parse_with_http_environment(config)
        config = self._parse_with_specific_args(config)
        mimeo_config = MimeoConfig(config)
        logger.fine(f"Parsed Mimeo Configuration: [{mimeo_config}]")
        return mimeo_config

    def _parse_with_http_environment(self, config: dict) -> dict:
        """Parse Mimeo Configuration with an HTTP environment.

        When --http-envs-file argument has not been provided, then
        the default environments path is used: mimeo.envs.json.

        Parameters
        ----------
        config : dict
            A source Mimeo Configuration dictionary

        Returns
        -------
        config : dict
            An overwritten Mimeo Configuration dictionary when
            --http-env argument has been provided. Otherwise,
            the source one, without any modification.

        Raises
        ------
        EnvironmentsFileNotFound
            If environments file does not exist.
        EnvironmentNotFound
            If the http environment is not defined in the environments file
        """
        if self._args.http_env is not None:
            envs_file_path = self._args.http_envs_file if self._args.http_envs_file is not None else DEFAULT_ENVS_FILE_PATH
            env = self._get_environment(envs_file_path, self._args.http_env)
            self._overwrite_output_details_with_env(config, env)
        return config

    def _parse_with_specific_args(self, config: dict) -> dict:
        """Parse Mimeo Configuration with Mimeo Configuration args.

        This method uses an internal `dict` fetching each command line
        argument with a corresponding Mimeo Configuration entry.

        Parameters
        ----------
        config : dict
            A source Mimeo Configuration dictionary

        Returns
        -------
        config : dict
            An overwritten Mimeo Configuration dictionary when
            some configuration arguments have been provided. Otherwise,
            the source one, without any modification.
        """
        for arg_name, mapping in self._ARGS_MAPPING.items():
            if hasattr(self._args, arg_name) and getattr(self._args, arg_name) is not None:
                entry_path = mapping.get(self._ENTRY_PATH_KEY, arg_name)
                get_value = mapping.get(self._GET_VALUE_KEY, lambda arg: arg)
                value = get_value(getattr(self._args, arg_name))
                config = self._overwrite_config_entry(config, entry_path, value)
        return config

    @classmethod
    def _overwrite_output_details_with_env(cls, config: dict, env: dict) -> dict:
        """Overwrite Mimeo Output Details with an environment config.

        Parameters
        ----------
        config : dict
            A source Mimeo Configuration dictionary
        env : dict
            An HTTP environment

        Returns
        -------
        config : dict
            An overwritten Mimeo Configuration.
        """
        for prop in cls._ENVIRONMENT_PROPS:
            value = env.get(prop)
            if value is not None:
                config_entry_path = [MimeoConfig.OUTPUT_DETAILS_KEY, prop]
                config = cls._overwrite_config_entry(config, config_entry_path, value)
        return config

    @classmethod
    def _overwrite_config_entry(cls, config_entry: dict, entry_path: list, value: Union[str, int, bool]) -> dict:
        """Overwrite a Mimeo Configuration entry.

        Recursively finds an entry using `entry_path` list. If any of
        middle entries does not exist, it is initialized as an empty
        dictionary. Once the target entry is found, the `value` is put
        there.

        Parameters
        ----------
        config_entry : dict
            A Mimeo Configuration entry to get its direct child
        entry_path : list
            A list of Mimeo Configuration nested properties, e.g.
            ['output_details', 'direction'] points to the direction
            property under output details entry.
        value
            An overwriting value when `entry_path` contains only one
            element.

        Returns
        -------
        config_entry : dict
            An overwritten Mimeo Configuration entry.
        """
        direct_entry = entry_path[0]
        if len(entry_path) > 1:
            if config_entry.get(direct_entry) is None:
                config_entry[direct_entry] = {}
            config_entry[direct_entry] = cls._overwrite_config_entry(config_entry[direct_entry], entry_path[1:], value)
        else:
            logger.fine(f"Overwriting {direct_entry} to [{value}]")
            config_entry[direct_entry] = value
        return config_entry

    @staticmethod
    def _get_environment(envs_path: str, env_name: str) -> dict:
        """Load an HTTP environment from HTTP envs vile.

        Parameters
        ----------
        envs_path : str
            An HTTP envs file path
        env_name : str
            An HTTP environment name

        Returns
        -------
        env : dict
            An HTTP environment configuration

        Raises
        ------
        EnvironmentsFileNotFound
            If environments file does not exist.
        EnvironmentNotFound
            If the http environment is not defined in the environments file
        """
        if path.exists(envs_path):
            with open(envs_path) as envs_file:
                envs = json.load(envs_file)
                if env_name in envs:
                    env = envs[env_name]
                    logger.debug(f"Using environment [{env_name}] from file [{envs_path}]: [{env}]")
                    return env
                else:
                    raise EnvironmentNotFound(env_name, envs_path)
        else:
            raise EnvironmentsFileNotFound(envs_path)



"""The Mimeo Meta Exceptions module.

It contains all custom exceptions related to useful abstract classes:
    * InstanceNotAlive
        A custom Exception class for using non-alive instance.
"""


class InstanceNotAlive(Exception):
    """A custom Exception class for using non-alive instance.

    Raised when using OnlyOneAlive class that is not alive.
    """

    def __init__(self):
        """Initialize InstanceNotAlive exception with details.

        Extends Exception constructor with a constant message.
        """
        super().__init__("The instance is not alive!")



"""The Mimeo XML Generator module.

It exports only one class:
    * XMLGenerator
        A Generator implementation producing data in the XML format.
"""
import logging
import xml.etree.ElementTree as ElemTree
from typing import Iterator, List, Union
from xml.dom import minidom

from mimeo.config.mimeo_config import MimeoConfig, MimeoTemplate
from mimeo.context import MimeoContext
from mimeo.context.decorators import (mimeo_clear_iterations, mimeo_context,
                                      mimeo_context_switch,
                                      mimeo_next_iteration)
from mimeo.generators import Generator
from mimeo.utils import MimeoRenderer

logger = logging.getLogger(__name__)


class XMLGenerator(Generator):
    """A Generator implementation producing data in the XML format.

    This Generator is instantiated for the 'xml' output format
    and produces data using Mimeo Configuration.

    Methods
    -------
    generate(
        templates: Union[list, Iterator[MimeoTemplate]],
        parent: ElemTree.Element = None
    ) -> Iterator[ElemTree.Element]
        Generate XML data based on the Mimeo Configuration.
    stringify(
        data: Any,
        mimeo_config: MimeoConfig
    ) -> str
        Stringify data generated by the generate() method.
    """

    def __init__(
            self,
            mimeo_config: MimeoConfig
    ):
        """Initialize XMLGenerator class.

        Parameters
        ----------
        mimeo_config : MimeoConfig
            A Mimeo Configuration
        """
        self.__indent = mimeo_config.output_details.indent
        self.__xml_declaration = mimeo_config.output_details.xml_declaration

    @classmethod
    def generate(
            cls,
            templates: Union[list, Iterator[MimeoTemplate]],
            parent: ElemTree.Element = None
    ) -> Iterator[ElemTree.Element]:
        """Generate XML data based on the Mimeo Configuration.

        This function is used recursively when a Mimeo Configuration
        contains nested templates.
        It iterates through all templates configured and yields data
        units.

        Parameters
        ----------
        templates : Union[list, Iterator[MimeoTemplate]]
            A collection of Mimeo Templates to process
        parent : ElemTree.Element, default None
            A parent XML node for the currently processed template.
            It is passed only when a Mimeo Config contain nested
            templates.

        Returns
        -------
        Iterator[ElemTree.Element]
            Iterator for generated nodes
        """
        for template in templates:
            for data_unit in cls._process_single_template(template, parent):
                yield data_unit

    def stringify(
            self,
            data_unit: ElemTree.Element,
            mimeo_config: MimeoConfig
    ) -> str:
        """Stringify XML data generated by the generate() method.

        Parameters
        ----------
        data_unit: ElemTree.Element
            A single data unit generated by the generate() method
        mimeo_config: MimeoConfig
            A Mimeo Configuration providing output details

        Returns
        -------
        str
            Stringified data unit
        """
        if self.__indent is None or self.__indent == 0:
            return ElemTree.tostring(data_unit,
                                     encoding="utf-8",
                                     method="xml",
                                     xml_declaration=self.__xml_declaration).decode('ascii')
        else:
            xml_string = ElemTree.tostring(data_unit)
            xml_minidom = minidom.parseString(xml_string)
            if self.__xml_declaration:
                return xml_minidom.toprettyxml(indent=" " * self.__indent, encoding="utf-8").decode('ascii')
            else:
                return xml_minidom.childNodes[0].toprettyxml(indent=" " * self.__indent, encoding="utf-8").decode('ascii')

    @classmethod
    @mimeo_context_switch
    @mimeo_clear_iterations
    def _process_single_template(
            cls,
            template: MimeoTemplate,
            parent: ElemTree.Element = None
    ) -> List[ElemTree.Element]:
        """Process a single Mimeo Template.

        This function is used recursively when a Mimeo Configuration
        contains nested templates.
        It repeats same processing operation so many times as it is
        configured in the `count` property of a Mimeo Configuration.
        Before the template execution it switches context managed by
        MimeoContextManager and also clears iterations as nested
        templates would collect iterations from previous parent's
        iterations.

        Parameters
        ----------
        template : MimeoTemplate
            A single Mimeo Template to process
        parent : ElemTree.Element, default None
            A parent node for processing nested templates

        Returns
        -------
        List[ElemTree.Element]
            A list of generated data units
        """
        logger.debug(f"Reading template [{template}]")
        data_units = [cls._process_single_data_unit(template, parent) for _ in iter(range(template.count))]
        return data_units

    @classmethod
    @mimeo_next_iteration
    def _process_single_data_unit(
            cls,
            template: MimeoTemplate,
            parent: ElemTree.Element = None
    ) -> ElemTree.Element:
        """Process a single data unit from the template.

        This function is used recursively when a Mimeo Configuration
        contains nested templates.
        It processes a single data unit. The reason why it is separated
        from the _process_node() function is the @mimeo_next_iteration
        decorator. The _process_node() function is recursively called
        by itself, and it produces data for a single iteration.
        The purpose of this function is to increment iteration before
        processing node.

        Parameters
        ----------
        template : MimeoTemplate
            A single Mimeo Template to process
        parent : ElemTree.Element, default None
            A parent node for processing nested templates

        Returns
        -------
        ElemTree.Element
            A single data unit generated within a single template
            iteration. If the `parent` is not None it will not return
            any value.
        """
        return cls._process_node(parent, template.model.root_name, template.model.root_data)

    @classmethod
    @mimeo_context
    def _process_node(
            cls,
            parent: ElemTree.Element,
            element_tag: str,
            element_value: Union[dict, list, str, int, float, bool],
            attributes: dict = None,
            context: MimeoContext = None
    ) -> ElemTree.Element:
        """Process a single template's node.

        This is a recursive function that traverses Mimeo Template
        and generates XML nodes based on `element_value` type.

        Parameters
        ----------
        parent : ElemTree.Element
            A parent node
        element_tag : str
            An element tag
        element_value : Union[dict, list, str, int, float, bool]
            An element value
        attributes : dict, default None
            Element's attributes
        context : MimeoContext, default None
            The current Mimeo Context (injected by MimeoContextManager)

        Returns
        -------
        ElemTree.Element
            Returned only when `parent` is None. A single data unit
            generated within a single template iteration.

        Raises
        ------
        InvalidSpecialFieldValue
            If the special field value is dict or list
        SpecialFieldNotFound
            If the special field does not exist.
        """
        logger.fine(f"Rendering element - "
                    f"parent [{parent if parent is None else parent.tag}], "
                    f"element_tag [{element_tag}], "
                    f"element_value [{element_value}], "
                    f"attributes [{attributes}]")
        attributes = attributes if attributes is not None else {}
        if element_tag == MimeoConfig.TEMPLATES_KEY:
            templates = (MimeoTemplate(template) for template in element_value)
            for _ in cls.generate(templates, parent):
                pass
        else:
            is_special_field = MimeoRenderer.is_special_field(element_tag)
            if is_special_field:
                element_tag = MimeoRenderer.get_special_field_name(element_tag)

            element = cls._create_xml_element(parent, element_tag, attributes)

            if isinstance(element_value, dict) and not MimeoRenderer.is_parametrized_mimeo_util(element_value):
                if MimeoConfig.MODEL_ATTRIBUTES_KEY in element_value:
                    element_value_copy = dict(element_value)
                    attrs = element_value_copy.pop(MimeoConfig.MODEL_ATTRIBUTES_KEY)
                    value = element_value_copy.get(MimeoConfig.MODEL_VALUE_KEY, element_value_copy)
                    if parent is not None:
                        parent.remove(element)
                        cls._process_node(parent, element_tag, value, attrs)
                    else:
                        return cls._process_node(parent, element_tag, value, attrs)
                else:
                    for child_tag, child_value in element_value.items():
                        cls._process_node(element, child_tag, child_value)
            elif isinstance(element_value, list):
                has_only_atomic_values = all(not isinstance(child, (list, dict)) for child in element_value)
                if has_only_atomic_values:
                    parent.remove(element)
                    for child in element_value:
                        cls._process_node(parent, element_tag, child)
                else:
                    for child in element_value:
                        grand_child_tag = next(iter(child))
                        grand_child_data = child[grand_child_tag]
                        cls._process_node(element, grand_child_tag, grand_child_data)
            else:
                value = MimeoRenderer.render(element_value)
                if is_special_field:
                    context.curr_iteration().add_special_field(element_tag, value)

                value_str = str(value) if value is not None else ""
                element.text = value_str.lower() if isinstance(value, bool) else value_str
                logger.fine(f"Rendered value [{element.text}]")

            if parent is None:
                return element

    @classmethod
    def _create_xml_element(
            cls,
            parent: ElemTree.Element,
            element_tag: str,
            attributes: dict
    ) -> Union[ElemTree.Element, ElemTree.SubElement]:
        """Create an XML element based on the `parent` existence.

        Parameters
        ----------
        parent : ElemTree.Element
            A parent node
        element_tag : str
            An element tag
        attributes : dict
            Element's attributes

        Returns
        -------
        Union[ElemTree.Element, ElemTree.SubElement]
            If the `parent` is None, returns ElemTree.Element.
            Otherwise, returns ElemTree.SubElement
        """
        if parent is None:
            return ElemTree.Element(element_tag, attrib=attributes)
        else:
            return ElemTree.SubElement(parent, element_tag, attrib=attributes)



"""The Mimeo Configuration Exceptions module.

It contains all custom exceptions related to Mimeo Configuration:
    * UnsupportedPropertyValue
        A custom Exception class for unsupported properties' values.
    * MissingRequiredProperty
        A custom Exception class for missing required properties.
    * InvalidIndent
        A custom Exception class for invalid indent configuration.
    * InvalidVars
        A custom Exception class for invalid vars' configuration.
    * InvalidMimeoModel
        A custom Exception class for invalid model configuration.
    * InvalidMimeoTemplate
        A custom Exception class for invalid template configuration.
    * InvalidMimeoConfig
        A custom Exception class for invalid mimeo configuration.
"""


class UnsupportedPropertyValue(Exception):
    """A custom Exception class for unsupported properties' values.

    Raised when a Mimeo Configuration property points to a value
    not being supported by Mimeo.
    """

    def __init__(self, prop: str, val: str, supported_values: tuple):
        """Initialize UnsupportedPropertyValue exception with details.

        Extends Exception constructor with a custom message.

        Parameters
        ----------
        prop : str
            A property name
        val : str
            A property value
        supported_values : tuple
            A list of supported values for the property
        """
        super().__init__(f"Provided {prop} [{val}] is not supported! "
                         f"Supported values: [{', '.join(supported_values)}].")


class MissingRequiredProperty(Exception):
    """A custom Exception class for missing required properties.

    Raised when a Mimeo Configuration does not contain a required
    property.
    """

    pass


class InvalidIndent(Exception):
    """A custom Exception class for invalid indent configuration.

    Raised when a configured indent is negative.
    """

    def __init__(self, indent: int):
        """Initialize InvalidIndent exception with details.

        Extends Exception constructor with a custom message.

        Parameters
        ----------
        indent : int
            A configured indent
        """
        super().__init__(f"Provided indent [{indent}] is negative!")


class InvalidVars(Exception):
    """A custom Exception class for invalid vars' configuration.

    Raised when vars are not configured properly.
    """

    pass


class InvalidMimeoModel(Exception):
    """A custom Exception class for invalid model configuration.

    Raised when a Mimeo Model is not configured properly.
    """

    pass


class InvalidMimeoTemplate(Exception):
    """A custom Exception class for invalid template configuration.

    Raised when a Mimeo Template is not configured properly.
    """

    pass


class InvalidMimeoConfig(Exception):
    """A custom Exception class for invalid mimeo configuration.

    Raised when a Mimeo Configuration is not configured properly.
    """

    pass
